{
 Ideas:
  * use opparse.scm to parse arithmetic expressions into syntax trees
  * map syntax trees into constraint contractor objects
    (these are objects with several operations -- contract, getvars, setvars)
  * implement a GUI using contractors 
  * reimplement the plotter 
  * implement as applet, jws, application, groupware tool!

Steps:
 * modify opparse.scm so it returns a syntax tree
 * create a constraint contractor object from a syntax tree with operations:
    contract
    getBindings
    setBindings
    resetBindings
 * create a constraint solver that reads the constraint, creates a contractor,
   invokes it N times, and returns the result
 * wrap it in a Java class

}

(import "net.sourceforge.interval.ia_math.RealInterval")
(import "net.sourceforge.interval.ia_math.IANarrow")

(use-module "ia_solver/opparse.scm")

(define (strip-parens expr)
  (if (string? expr) expr
      (if (equal? (first expr) "()")
          (second expr)
          (append (list (first expr)) (map strip-parens (rest expr))))))

(define (testread x)
  (define p (makeparser arithexprs))
  (strip-parens (p x)))

 ; syntree x hashtable -> contractor
 ; syntree is a syntax tree created by a parser
 ; H is a hashtable associating intervals to variables

(define (solve expr N)
  (define s (make-solver expr))
  (let loop ((N N))
     (s 'contract)
     (if (> N 0) 
         (loop (- N 1))
         (s 'h))))


  (define (make-solver expr)
   (define p (testread expr))
   (define h (Hashtable.))
   (define k (make-contractor p h))
   (lambda (msg)
     (case msg
      ((h) h)
      ((k) k)
      ((p) p)
      ((contract) (k 'contract)))))


  (define (make-contractor syntree H)
;    (display (list 'makingcontractor syntree H))(newline)
     (if (string? syntree)
         (process-leaf syntree H)
         (process-expression syntree H)))

  ; string x hashtable -> constraint
  (define (process-leaf leaf H)
     (let ((z (string->number leaf)))
       (if z 
          (make-number leaf z)
          (make-variable leaf H))))



  (define (process-expression syntree H)
    (make-complex-contraint syntree (map (lambda(x) (make-contractor x H)) (rest syntree))))

  ; string -> constraint
  (define (make-variable syntree H) 
    (if (equal? (.get H syntree) #null)
        (.put H syntree (RealInterval.)))
    (let ((val (.get H syntree)))
      (lambda (msg)
        (case msg
         ((type) 'variable)
         ((syntax) syntree)
         ((val) val)
         ((contract) #t)))))

  ; number -> constraint
  (define (make-number syntree num) 
    (let* 
      ((dnum (.doubleValue num))
       (val (RealInterval. dnum dnum)))
      (lambda (msg)
        (case msg
         ((type) 'number)
         ((syntax) syntree)
         ((val) val)
         ((contract) #t)))))

  ; syntree x args -> contractor
  (define  (make-complex-contraint syntree args)
    (let* ((op (first syntree))
           (val (RealInterval.))
           (vals (map (lambda(x) (x 'val)) args))
           (contract-me (lambda() (contract op val vals)))
           (contract-children (lambda() (for-each (lambda(x) (x 'contract)) args)))
         )
      (lambda(msg)
        (case msg
         ((type) 'expression)
         ((syntax) syntree)
         ((val) val)
         ((contract) 
          (and 
             (contract-me)
             (contract-children)
             (contract-me)))
      ))))



(define ONE (RealInterval. 1.0 1.0))

(define (contract op val args)
    (do-contract op val args))
;    (debug-contract op val args))


(define (debug-contract op val args)
  (display (list 'contracting op val args))(newline)
  (let ((v (do-contract op val args)))
     (display (list 'contracted op val args 'toget v)) (newline)(newline)
     v))

(define (do-contract op val  args)
  (case op
     (("+") (IANarrow.narrow_add val (first args) (second args) ))
     (("-") (if (equal? 2 (length args))
                (IANarrow.narrow_sub val (first args) (second args) )
                (IANarrow.narrow_uminus val (first args))))
     (("*") (IANarrow.narrow_mul val (first args) (second args)))
     (("/") (IANarrow.narrow_div val (first args) (second args)))
     (("exp") (IANarrow.narrow_exp    (first args) val))
     (("log") (IANarrow.narrow_log    (first args) val))
     (("sin") (IANarrow.narrow_sin    (first args) val))
     (("cos") (IANarrow.narrow_cos    (first args) val))
     (("tan") (IANarrow.narrow_tan    (first args) val))
     (("asin") (IANarrow.narrow_asin  (first args) val))
     (("acos") (IANarrow.narrow_acos  (first args) val))
     (("atan") (IANarrow.narrow_atan  (first args) val))
     (("^") (IANarrow.narrow_carot val (first args) (second args)))
     (("**") (IANarrow.narrow_power val (first args) (second args)))
     ((";") (IANarrow.narrow_semi val (first args) (second args)))
     ((":=") (IANarrow.narrow_colon_equals val (first args) (second args)))
     (("=") (IANarrow.narrow_equals (first args) (second args)))
     (("==") (IANarrow.narrow_eq val (first args) (second args) (third args)))
     (("<") (IANarrow.narrow_lt ONE (first args) (second args)))
     (("<=") (IANarrow.narrow_le ONE (first args) (second args)))
     ((">") (IANarrow.narrow_gt ONE (first args) (second args)))
     ((">=") (IANarrow.narrow_ge ONE (first args) (second args)))
     (("<>") (IANarrow.narrow_ne ONE (first args) (second args)))
     (else (display {error in ia-solver.scm: unknown operator [op]\n}))
    ))

